from setuptools import setup

setup(
    name='muse_sc',
    version='0.0.4',
    description='MUSE is a deep learning approach characterizing tissue composition through combined analysis of morphologies and transcriptional states for single-cell spatial transcriptomics data.',
    license='MIT',
    packages=['muse_sc'],
    author='Feng Bao @ AltschulerWu-Lab',
    author_email='feng.bao@ucsf.edu',
    keywords=['single-cell analysis', 'deep learning', 'spatial transcriptomics'],
    url='https://github.com/AltschulerWu-Lab/MUSE'
)
